package ij.plugin.filter;
import ij.*;
import ij.gui.*;
import ij.process.*;
import ij.measure.*;
import ij.util.*;
import java.awt.*;
import java.util.*;

/** Implements the Analyze/Calibrate command. */
public class Calibrator implements PlugInFilter, Measurements {

	private static final String NONE = "None";
	private static final String INVERTER = "Pixel Inverter";
	private static final String UNCALIBRATED_OD = "Uncalibrated OD";
	static boolean global;
	private static boolean oldGlobal;
    private ImagePlus imp;
	private int choiceIndex;
	private String[] functions;
	private	int nFits = CurveFitter.fitList.length;
	private int spacerIndex = nFits+1;
	private int inverterIndex = nFits+2;
	private int odIndex = nFits+3;
	private String xText;
	private static String yText="";
	private String valueUnit;

	public int setup(String arg, ImagePlus imp) {
		this.imp = imp;
		IJ.register(Calibrator.class);
		return DOES_8G+NO_CHANGES;
	}

	public void run(ImageProcessor ip) {
		oldGlobal = global;
		if (!showDialog(imp))
			return;
		calibrate(imp);
		if (global || global!=oldGlobal) {
			int[] list = WindowManager.getIDList();
			if (list==null)
				return;
			for (int i=0; i<list.length; i++) {
				ImagePlus imp2 = WindowManager.getImage(list[i]);
				if (imp2!=null)
					imp2.getWindow().repaint();
			}
		} else
			imp.getWindow().repaint();
	}

	public boolean showDialog(ImagePlus imp) {
		String defaultChoice;
		Calibration cal = imp.getCalibration();
		functions = getFunctionList();
		int function = cal.getFunction();
		valueUnit = cal.getValueUnit();
		if (function==Calibration.NONE)
			defaultChoice=NONE;
		else if (function<nFits)
			defaultChoice = CurveFitter.fitList[function];
		else if (function==Calibration.INVERTER)
			defaultChoice=INVERTER;
		else if (function==Calibration.UNCALIBRATED_OD)
			defaultChoice=UNCALIBRATED_OD;
		else
			defaultChoice=NONE;
			
		xText = getMeans();	
		GenericDialog gd = new GenericDialog("Calibrate...", IJ.getInstance());
		gd.addChoice("Function:", functions, defaultChoice);
		gd.addStringField("Unit:", valueUnit, 16);
		gd.addTextAreas(xText, yText, 12, 10);
		gd.addCheckbox("Global Calibration", global);
		gd.showDialog();
		if (gd.wasCanceled())
			return false;
		else {
			choiceIndex = gd.getNextChoiceIndex();
			valueUnit = gd.getNextString();
			xText = gd.getNextText();
			yText = gd.getNextText();
			global = gd.getNextBoolean();
			return true;
		}
	}

	public void calibrate(ImagePlus imp) {
		Calibration cal = imp.getCalibration();
		int function = Calibration.NONE;
		float[] table = null;
		if (choiceIndex<=0) {
			if (!yText.equals(""))
				IJ.showMessage("Calibrator", "Please select a function");
			function = Calibration.NONE;
		} else if (choiceIndex<=nFits) {
			function = choiceIndex - 1;
			table = doCurveFitting(function);
		} else if (choiceIndex==inverterIndex) {
			table = new float[256];
			for (int i=0; i<256; i++)
				table[i] = 255-i;
			function = Calibration.INVERTER;
		} else if (choiceIndex==odIndex) {
			table = new float[256];
			boolean invertedLut = imp.isInvertedLut();
			for (int i=0; i<256; i++)
				table[i] = (float)od(i,invertedLut);
			function = Calibration.UNCALIBRATED_OD;
		}
		cal.setCTable(table, function);
		cal.setValueUnit(valueUnit);
		if (global)
			imp.setGlobalCalibration(cal);
		else
			imp.setCalibration(cal);
		//IJ.write("cal: "+cal);
	}

	double od(double v, boolean invertedLut) {
		if (invertedLut) {
			if (v==255.0) v = 254.5;
			return 0.434294481*Math.log(255.0/(255.0-v));
		} else {
			if (v==0.0) v = 0.5;
			return 0.434294481*Math.log(255.0/v);
		}
	}
	
	float[] doCurveFitting(int fitType) {
		double[] x = getData(xText);
		double[] y = getData(yText);
		if (x.length!=y.length || y.length==0) {
			IJ.showMessage("Calibrator",
				"To create a calibration curve, the left column must\n"+
				"contain a list of measured mean pixel values and the\n"+
				"right column must contain the same number of calibration\n"+
				"standard values. Use the Measure command to add mean\n"+
				"pixel value measurements to the left column.\n"
				);
			return null;
		}
		int n = x.length;
		double[] a;
		//double[] a = Tools.getMinMax(x);
		//double xmin=a[0], xmax=a[1]; 
		double xmin=0.0, xmax=255.0; 
		a = Tools.getMinMax(y);
		double ymin=a[0], ymax=a[1]; 
		CurveFitter cf = new CurveFitter(x, y);
		cf.doFit(fitType);
		IJ.write("");
		IJ.write("n: "+n);
		IJ.write("iterations: "+cf.getIterations());
		IJ.write("max iterations: "+cf.getMaxIterations());
		IJ.write("function: "+cf.fList[fitType]);
		int nc = cf.nCoefficients();
		double[] c = cf.getCoefficients();
		IJ.write("a: "+IJ.d2s(c[0],6));
		IJ.write("b: "+IJ.d2s(c[1],6));
		if (nc>=3)
			IJ.write("c: "+IJ.d2s(c[2],6));
		if (nc>=4)
			IJ.write("d: "+IJ.d2s(c[3],6));
		if (nc>=5)
			IJ.write("e: "+IJ.d2s(c[4],6));
		double sumResidualsSqr = c[nc];
		IJ.write("sum of residuals: "+IJ.d2s(Math.sqrt(sumResidualsSqr),6));
		double sumY = 0.0;
		for (int i=0; i<n; i++)
			sumY += y[i];
		double sd = Math.sqrt(sumResidualsSqr/n);
		double mean = sumY/n;
		double sumMeanDiffSqr = 0.0;
		int degreesOfFreedom = n-nc;
		double fitGoodness=1.0;
		for (int i=0; i<n; i++) {
			sumMeanDiffSqr += sqr(y[i]-mean);
			if (sumMeanDiffSqr>0.0 && degreesOfFreedom!=0)
				fitGoodness = 1.0-(sumResidualsSqr/degreesOfFreedom)*((n-1)/sumMeanDiffSqr);
		}
		IJ.write("S.D.: "+IJ.d2s(sd,6));
		IJ.write("R^: "+IJ.d2s(fitGoodness,6));

		int range = 256;
		float[] px = new float[range];
		float[] py = new float[range];
		double inc = (xmax-xmin)/(range-1);
		double tmp = xmin;
		for (int i=0; i<range; i++) {
			px[i]=(float)tmp;
			tmp += inc;
		}
		for (int i=0; i<range; i++)
			py[i] = (float)cf.f(c, px[i]);
		a = Tools.getMinMax(py);
		ymin = Math.min(ymin, a[0]);
		ymax = Math.max(ymax, a[1]);
		PlotWindow pw = new PlotWindow(cf.fList[fitType],"pixel value",valueUnit,px,py);
		pw.setLimits(xmin,xmax,ymin,ymax);
		pw.addPoints(x, y, PlotWindow.CIRCLE);
		//pw.addLabel(0.02, 0.1, cf.fList[fitType]);
		pw.draw();
		return py;									
	}
	
	double sqr(double x) {return x*x;}

	String[] getFunctionList() {
		String[] list = new String[nFits+4];
		list[0] = NONE;
		for (int i=0; i<nFits; i++)
			list[1+i] = CurveFitter.fitList[i];
		list[spacerIndex] = "-";
		list[inverterIndex] = INVERTER;
		list[odIndex] = UNCALIBRATED_OD;
		return list;
 	}
	
	String getMeans() {
		float[] umeans = Analyzer.getUMeans();
		int count = Analyzer.getCounter();
		if (umeans==null || count==0)
			return "";
		if (count>MAX_STANDARDS)
			count = MAX_STANDARDS;
		String s = "";
		for (int i=0; i<count; i++)
			s += IJ.d2s(umeans[i],2)+"\n";
		return s;
	}

	double[] getData(String xData) {
		StringTokenizer st = new StringTokenizer(xData);
		int nTokens = st.countTokens();
		if (nTokens<1)
			return new double[0];
		int n = nTokens;
		double data[] = new double[n];
		for (int i=0; i<n; i++) {
			data[i] = getNum(st);
		}
		return data;
	}
	
	double getNum(StringTokenizer st) {
		Double d;
		String token = st.nextToken();
		try {d = new Double(token);}
		catch (NumberFormatException e){d = null;}
		if (d!=null)
			return(d.doubleValue());
		else
			return 0.0;
	}

}